from typing import List, Tuple, Dict, Union
import warnings

import mujoco
from mujoco import MjSpec

import loco_mujoco
from loco_mujoco.environments.humanoids.base_skeleton import BaseSkeleton


class SkeletonTorque(BaseSkeleton):

    """
    Description
    ------------

    Mujoco environment of a humanoid skeleton model with one torque actuator per joint.


    Default Observation Space
    -----------------

    ============ =================== ================ ==================================== ============================== ===
    Index in Obs Name                ObservationType  Min                                  Max                            Dim
    ============ =================== ================ ==================================== ============================== ===
    0 - 4        q_root              FreeJointPosNoXY [-inf, -inf, -inf, -inf, -inf]       [inf, inf, inf, inf, inf]      5
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    5            q_hip_flexion_r     JointPos         [-0.787]                             [0.787]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    6            q_hip_adduction_r   JointPos         [-0.524]                             [0.524]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    7            q_hip_rotation_r    JointPos         [-2.0944]                            [2.0944]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    8            q_knee_angle_r      JointPos         [-2.0944]                            [0.174533]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    9            q_ankle_angle_r     JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    10           q_hip_flexion_l     JointPos         [-0.787]                             [0.787]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    11           q_hip_adduction_l   JointPos         [-0.524]                             [0.524]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    12           q_hip_rotation_l    JointPos         [-2.0944]                            [2.0944]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    13           q_knee_angle_l      JointPos         [-2.0944]                            [0.174533]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    14           q_ankle_angle_l     JointPos         [-1.0472]                            [1.0472]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    15           q_lumbar_extension  JointPos         [-1.5708]                            [0.377]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    16           q_lumbar_bending    JointPos         [-0.754]                             [0.754]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    17           q_lumbar_rotation   JointPos         [-0.754]                             [0.754]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    18           q_arm_flex_r        JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    19           q_arm_add_r         JointPos         [-2.0944]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    20           q_arm_rot_r         JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    21           q_elbow_flex_r      JointPos         [-0.3]                               [2.618]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    22           q_pro_sup_r         JointPos         [-0.6]                               [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    23           q_wrist_flex_r      JointPos         [-1.22173]                           [1.22173]                      1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    24           q_wrist_dev_r       JointPos         [-0.436332]                          [0.610865]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    25           q_arm_flex_l        JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    26           q_arm_add_l         JointPos         [-2.0944]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    27           q_arm_rot_l         JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    28           q_elbow_flex_l      JointPos         [-0.3]                               [2.618]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    29           q_pro_sup_l         JointPos         [-0.6]                               [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    30           q_wrist_flex_l      JointPos         [-1.22173]                           [1.22173]                      1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    31           q_wrist_dev_l       JointPos         [-0.436332]                          [0.610865]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    32 - 37      dq_root             FreeJointVel     [-inf, -inf, -inf, -inf, -inf, -inf] [inf, inf, inf, inf, inf, inf] 6
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    38           dq_hip_flexion_r    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    39           dq_hip_adduction_r  JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    40           dq_hip_rotation_r   JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    41           dq_knee_angle_r     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    42           dq_ankle_angle_r    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    43           dq_hip_flexion_l    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    44           dq_hip_adduction_l  JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    45           dq_hip_rotation_l   JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    46           dq_knee_angle_l     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    47           dq_ankle_angle_l    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    48           dq_lumbar_extension JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    49           dq_lumbar_bending   JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    50           dq_lumbar_rotation  JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    51           dq_arm_flex_r       JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    52           dq_arm_add_r        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    53           dq_arm_rot_r        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    54           dq_elbow_flex_r     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    55           dq_pro_sup_r        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    56           dq_wrist_flex_r     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    57           dq_wrist_dev_r      JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    58           dq_arm_flex_l       JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    59           dq_arm_add_l        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    60           dq_arm_rot_l        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    61           dq_elbow_flex_l     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    62           dq_pro_sup_l        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    63           dq_wrist_flex_l     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    64           dq_wrist_dev_l      JointVel         [-inf]                               [inf]                          1
    ============ =================== ================ ==================================== ============================== ===

    Default Action Space
    -------------------

    Control function type: **DefaultControl**

    See control function interface for more details.

    =============== ==== ===
    Index in Action Min  Max
    =============== ==== ===
    0               -1.0 1.0
    --------------- ---- ---
    1               -1.0 1.0
    --------------- ---- ---
    2               -1.0 1.0
    --------------- ---- ---
    3               -1.0 1.0
    --------------- ---- ---
    4               -1.0 1.0
    --------------- ---- ---
    5               -1.0 1.0
    --------------- ---- ---
    6               -1.0 1.0
    --------------- ---- ---
    7               -1.0 1.0
    --------------- ---- ---
    8               -1.0 1.0
    --------------- ---- ---
    9               -1.0 1.0
    --------------- ---- ---
    10              -1.0 1.0
    --------------- ---- ---
    11              -1.0 1.0
    --------------- ---- ---
    12              -1.0 1.0
    --------------- ---- ---
    13              -1.0 1.0
    --------------- ---- ---
    14              -1.0 1.0
    --------------- ---- ---
    15              -1.0 1.0
    --------------- ---- ---
    16              -1.0 1.0
    --------------- ---- ---
    17              -1.0 1.0
    --------------- ---- ---
    18              -1.0 1.0
    --------------- ---- ---
    19              -1.0 1.0
    --------------- ---- ---
    20              -1.0 1.0
    --------------- ---- ---
    21              -1.0 1.0
    --------------- ---- ---
    22              -1.0 1.0
    --------------- ---- ---
    23              -1.0 1.0
    --------------- ---- ---
    24              -1.0 1.0
    --------------- ---- ---
    25              -1.0 1.0
    --------------- ---- ---
    26              -1.0 1.0
    =============== ==== ===

    Methods
    ------------

    """
    mjx_enabled = False

    def __init__(self, **kwargs) -> None:
        """
        Constructor.

        """

        if "use_muscles" in kwargs.keys():
            assert kwargs["use_muscles"] is False, "Activating muscles in this environment not allowed. "
            del kwargs["use_muscles"]

        super(SkeletonTorque, self).__init__(use_muscles=False, **kwargs)

    @staticmethod
    def _get_action_specification(spec: MjSpec) -> List[str]:
        """
        Getter for the action space specification.

        Args:
            spec (MjSpec): Specification of the environment.

        Returns:
            List[str]: List of action space specification.
        """

        action_spec = ["mot_lumbar_ext", "mot_lumbar_bend", "mot_lumbar_rot", "mot_shoulder_flex_r",
                       "mot_shoulder_add_r", "mot_shoulder_rot_r", "mot_elbow_flex_r", "mot_pro_sup_r",
                       "mot_wrist_flex_r", "mot_wrist_dev_r", "mot_shoulder_flex_l", "mot_shoulder_add_l",
                       "mot_shoulder_rot_l", "mot_elbow_flex_l", "mot_pro_sup_l", "mot_wrist_flex_l",
                       "mot_wrist_dev_l", "mot_hip_flexion_r", "mot_hip_adduction_r", "mot_hip_rotation_r",
                       "mot_knee_angle_r", "mot_ankle_angle_r", "mot_subtalar_angle_r", "mot_mtp_angle_r",
                       "mot_hip_flexion_l", "mot_hip_adduction_l", "mot_hip_rotation_l", "mot_knee_angle_l",
                       "mot_ankle_angle_l", "mot_subtalar_angle_l", "mot_mtp_angle_l"]

        return action_spec

    @classmethod
    def get_default_xml_file_path(cls) -> str:
        """
        Returns the default path to the xml file of the environment.
        """
        return (loco_mujoco.PATH_TO_MODELS / "skeleton" / "skeleton_torque.xml").as_posix()


class MjxSkeletonTorque(SkeletonTorque):

    """
    Mjx version of SkeletonTorque.

    """

    mjx_enabled = True

    def __init__(self, timestep: float = 0.002,
                 n_substeps: int = 5,
                 **kwargs) -> None:
        """
        Constructor.

        Args:
            timestep (float): Timestep of the simulation.
            n_substeps (int): Number of substeps.
            **kwargs: Additional arguments
        """
        if "model_option_conf" not in kwargs.keys():
            model_option_conf = dict(iterations=4, ls_iterations=8, disableflags=mujoco.mjtDisableBit.mjDSBL_EULERDAMP)
        else:
            model_option_conf = kwargs["model_option_conf"]
            del kwargs["model_option_conf"]
        super().__init__(timestep=timestep, n_substeps=n_substeps, model_option_conf=model_option_conf, **kwargs)


class HumanoidTorque(SkeletonTorque):
    """
    Wrapper class for SkeletonTorque. Deprecated and will be removed in a future release.
    """
    def __init__(self, *args, **kwargs):
        warnings.warn(
            f"{self.__class__.__name__} is deprecated and will be removed in a future release. "
            f"Please use {super().__class__.__name__} instead.",
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*args, **kwargs)


class SkeletonMuscle(BaseSkeleton):

    """
    Description
    ------------

    Mujoco environment of a human skeleton model with muscle actuation.

    .. note:: This Humanoid consists of 92 muscles on the lower limb. The upper body is torque actuated.

    Default Observation Space
    -----------------

    ============ =================== ================ ==================================== ============================== ===
    Index in Obs Name                ObservationType  Min                                  Max                            Dim
    ============ =================== ================ ==================================== ============================== ===
    0 - 4        q_root              FreeJointPosNoXY [-inf, -inf, -inf, -inf, -inf]       [inf, inf, inf, inf, inf]      5
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    5            q_hip_flexion_r     JointPos         [-0.787]                             [0.787]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    6            q_hip_adduction_r   JointPos         [-0.524]                             [0.524]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    7            q_hip_rotation_r    JointPos         [-2.0944]                            [2.0944]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    8            q_knee_angle_r      JointPos         [-2.0944]                            [0.174533]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    9            q_ankle_angle_r     JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    10           q_hip_flexion_l     JointPos         [-0.787]                             [0.787]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    11           q_hip_adduction_l   JointPos         [-0.524]                             [0.524]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    12           q_hip_rotation_l    JointPos         [-2.0944]                            [2.0944]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    13           q_knee_angle_l      JointPos         [-2.0944]                            [0.174533]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    14           q_ankle_angle_l     JointPos         [-1.0472]                            [1.0472]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    15           q_lumbar_extension  JointPos         [-1.5708]                            [0.377]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    16           q_lumbar_bending    JointPos         [-0.754]                             [0.754]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    17           q_lumbar_rotation   JointPos         [-0.754]                             [0.754]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    18           q_arm_flex_r        JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    19           q_arm_add_r         JointPos         [-2.0944]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    20           q_arm_rot_r         JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    21           q_elbow_flex_r      JointPos         [0.0]                                [2.618]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    22           q_pro_sup_r         JointPos         [0.0]                                [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    23           q_wrist_flex_r      JointPos         [-1.22173]                           [1.22173]                      1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    24           q_wrist_dev_r       JointPos         [-0.436332]                          [0.610865]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    25           q_arm_flex_l        JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    26           q_arm_add_l         JointPos         [-2.0944]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    27           q_arm_rot_l         JointPos         [-1.5708]                            [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    28           q_elbow_flex_l      JointPos         [0.0]                                [2.618]                        1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    29           q_pro_sup_l         JointPos         [0.0]                                [1.5708]                       1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    30           q_wrist_flex_l      JointPos         [-1.22173]                           [1.22173]                      1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    31           q_wrist_dev_l       JointPos         [-0.436332]                          [0.610865]                     1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    32 - 37      dq_root             FreeJointVel     [-inf, -inf, -inf, -inf, -inf, -inf] [inf, inf, inf, inf, inf, inf] 6
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    38           dq_hip_flexion_r    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    39           dq_hip_adduction_r  JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    40           dq_hip_rotation_r   JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    41           dq_knee_angle_r     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    42           dq_ankle_angle_r    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    43           dq_hip_flexion_l    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    44           dq_hip_adduction_l  JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    45           dq_hip_rotation_l   JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    46           dq_knee_angle_l     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    47           dq_ankle_angle_l    JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    48           dq_lumbar_extension JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    49           dq_lumbar_bending   JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    50           dq_lumbar_rotation  JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    51           dq_arm_flex_r       JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    52           dq_arm_add_r        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    53           dq_arm_rot_r        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    54           dq_elbow_flex_r     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    55           dq_pro_sup_r        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    56           dq_wrist_flex_r     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    57           dq_wrist_dev_r      JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    58           dq_arm_flex_l       JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    59           dq_arm_add_l        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    60           dq_arm_rot_l        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    61           dq_elbow_flex_l     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    62           dq_pro_sup_l        JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    63           dq_wrist_flex_l     JointVel         [-inf]                               [inf]                          1
    ------------ ------------------- ---------------- ------------------------------------ ------------------------------ ---
    64           dq_wrist_dev_l      JointVel         [-inf]                               [inf]                          1
    ============ =================== ================ ==================================== ============================== ===

    Default Action Space
    -------------------

    Control function type: **DefaultControl**

    See control function interface for more details.

    =============== ==== ===
    Index in Action Min  Max
    =============== ==== ===
    0               -1.0 1.0
    --------------- ---- ---
    1               -1.0 1.0
    --------------- ---- ---
    2               -1.0 1.0
    --------------- ---- ---
    3               -1.0 1.0
    --------------- ---- ---
    4               -1.0 1.0
    --------------- ---- ---
    5               -1.0 1.0
    --------------- ---- ---
    6               -1.0 1.0
    --------------- ---- ---
    7               -1.0 1.0
    --------------- ---- ---
    8               -1.0 1.0
    --------------- ---- ---
    9               -1.0 1.0
    --------------- ---- ---
    10              -1.0 1.0
    --------------- ---- ---
    11              -1.0 1.0
    --------------- ---- ---
    12              -1.0 1.0
    --------------- ---- ---
    13              -1.0 1.0
    --------------- ---- ---
    14              -1.0 1.0
    --------------- ---- ---
    15              -1.0 1.0
    --------------- ---- ---
    16              -1.0 1.0
    --------------- ---- ---
    17              -1.0 1.0
    --------------- ---- ---
    18              -1.0 1.0
    --------------- ---- ---
    19              -1.0 1.0
    --------------- ---- ---
    20              -1.0 1.0
    --------------- ---- ---
    21              -1.0 1.0
    --------------- ---- ---
    22              -1.0 1.0
    --------------- ---- ---
    23              -1.0 1.0
    --------------- ---- ---
    24              -1.0 1.0
    --------------- ---- ---
    25              -1.0 1.0
    --------------- ---- ---
    26              -1.0 1.0
    --------------- ---- ---
    27              -1.0 1.0
    --------------- ---- ---
    28              -1.0 1.0
    --------------- ---- ---
    29              -1.0 1.0
    --------------- ---- ---
    30              -1.0 1.0
    --------------- ---- ---
    31              -1.0 1.0
    --------------- ---- ---
    32              -1.0 1.0
    --------------- ---- ---
    33              -1.0 1.0
    --------------- ---- ---
    34              -1.0 1.0
    --------------- ---- ---
    35              -1.0 1.0
    --------------- ---- ---
    36              -1.0 1.0
    --------------- ---- ---
    37              -1.0 1.0
    --------------- ---- ---
    38              -1.0 1.0
    --------------- ---- ---
    39              -1.0 1.0
    --------------- ---- ---
    40              -1.0 1.0
    --------------- ---- ---
    41              -1.0 1.0
    --------------- ---- ---
    42              -1.0 1.0
    --------------- ---- ---
    43              -1.0 1.0
    --------------- ---- ---
    44              -1.0 1.0
    --------------- ---- ---
    45              -1.0 1.0
    --------------- ---- ---
    46              -1.0 1.0
    --------------- ---- ---
    47              -1.0 1.0
    --------------- ---- ---
    48              -1.0 1.0
    --------------- ---- ---
    49              -1.0 1.0
    --------------- ---- ---
    50              -1.0 1.0
    --------------- ---- ---
    51              -1.0 1.0
    --------------- ---- ---
    52              -1.0 1.0
    --------------- ---- ---
    53              -1.0 1.0
    --------------- ---- ---
    54              -1.0 1.0
    --------------- ---- ---
    55              -1.0 1.0
    --------------- ---- ---
    56              -1.0 1.0
    --------------- ---- ---
    57              -1.0 1.0
    --------------- ---- ---
    58              -1.0 1.0
    --------------- ---- ---
    59              -1.0 1.0
    --------------- ---- ---
    60              -1.0 1.0
    --------------- ---- ---
    61              -1.0 1.0
    --------------- ---- ---
    62              -1.0 1.0
    --------------- ---- ---
    63              -1.0 1.0
    --------------- ---- ---
    64              -1.0 1.0
    --------------- ---- ---
    65              -1.0 1.0
    --------------- ---- ---
    66              -1.0 1.0
    --------------- ---- ---
    67              -1.0 1.0
    --------------- ---- ---
    68              -1.0 1.0
    --------------- ---- ---
    69              -1.0 1.0
    --------------- ---- ---
    70              -1.0 1.0
    --------------- ---- ---
    71              -1.0 1.0
    --------------- ---- ---
    72              -1.0 1.0
    --------------- ---- ---
    73              -1.0 1.0
    --------------- ---- ---
    74              -1.0 1.0
    --------------- ---- ---
    75              -1.0 1.0
    --------------- ---- ---
    76              -1.0 1.0
    --------------- ---- ---
    77              -1.0 1.0
    --------------- ---- ---
    78              -1.0 1.0
    --------------- ---- ---
    79              -1.0 1.0
    --------------- ---- ---
    80              -1.0 1.0
    --------------- ---- ---
    81              -1.0 1.0
    --------------- ---- ---
    82              -1.0 1.0
    --------------- ---- ---
    83              -1.0 1.0
    --------------- ---- ---
    84              -1.0 1.0
    --------------- ---- ---
    85              -1.0 1.0
    --------------- ---- ---
    86              -1.0 1.0
    --------------- ---- ---
    87              -1.0 1.0
    --------------- ---- ---
    88              -1.0 1.0
    --------------- ---- ---
    89              -1.0 1.0
    --------------- ---- ---
    90              -1.0 1.0
    --------------- ---- ---
    91              -1.0 1.0
    --------------- ---- ---
    92              -1.0 1.0
    --------------- ---- ---
    93              -1.0 1.0
    --------------- ---- ---
    94              -1.0 1.0
    --------------- ---- ---
    95              -1.0 1.0
    --------------- ---- ---
    96              -1.0 1.0
    --------------- ---- ---
    97              -1.0 1.0
    --------------- ---- ---
    98              -1.0 1.0
    --------------- ---- ---
    99              -1.0 1.0
    --------------- ---- ---
    100             -1.0 1.0
    --------------- ---- ---
    101             -1.0 1.0
    --------------- ---- ---
    102             -1.0 1.0
    --------------- ---- ---
    103             -1.0 1.0
    --------------- ---- ---
    104             -1.0 1.0
    --------------- ---- ---
    105             -1.0 1.0
    =============== ==== ===

    Methods
    ------------

    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor.

        """

        if "use_muscles" in kwargs.keys():
            assert kwargs["use_muscles"] is True, "Activating torque actuators in this environment not allowed. "
            del kwargs["use_muscles"]

        super(SkeletonMuscle, self).__init__(use_muscles=True, **kwargs)

    @staticmethod
    def _get_action_specification(spec: MjSpec) -> List[str]:
        """
        Getter for the action space specification.

        Args:
            spec (MjSpec): Specification of the environment.

        Returns:
            List[str]: List of action space specification.
        """

        action_spec = ["mot_shoulder_flex_r", "mot_shoulder_add_r", "mot_shoulder_rot_r", "mot_elbow_flex_r",
                       "mot_pro_sup_r", "mot_wrist_flex_r", "mot_wrist_dev_r", "mot_shoulder_flex_l",
                       "mot_shoulder_add_l", "mot_shoulder_rot_l", "mot_elbow_flex_l", "mot_pro_sup_l",
                       "mot_wrist_flex_l", "mot_wrist_dev_l", "glut_med1_r", "glut_med2_r",
                       "glut_med3_r", "glut_min1_r", "glut_min2_r", "glut_min3_r", "semimem_r", "semiten_r",
                       "bifemlh_r", "bifemsh_r", "sar_r", "add_long_r", "add_brev_r", "add_mag1_r", "add_mag2_r",
                       "add_mag3_r", "tfl_r", "pect_r", "grac_r", "glut_max1_r", "glut_max2_r", "glut_max3_r",
                       "iliacus_r", "psoas_r", "quad_fem_r", "gem_r", "peri_r", "rect_fem_r", "vas_med_r",
                       "vas_int_r", "vas_lat_r", "med_gas_r", "lat_gas_r", "soleus_r", "tib_post_r",
                       "flex_dig_r", "flex_hal_r", "tib_ant_r", "per_brev_r", "per_long_r", "per_tert_r",
                       "ext_dig_r", "ext_hal_r", "glut_med1_l", "glut_med2_l", "glut_med3_l", "glut_min1_l",
                       "glut_min2_l", "glut_min3_l", "semimem_l", "semiten_l", "bifemlh_l", "bifemsh_l",
                       "sar_l", "add_long_l", "add_brev_l", "add_mag1_l", "add_mag2_l", "add_mag3_l",
                       "tfl_l", "pect_l", "grac_l", "glut_max1_l", "glut_max2_l", "glut_max3_l",
                       "iliacus_l", "psoas_l", "quad_fem_l", "gem_l", "peri_l", "rect_fem_l",
                       "vas_med_l", "vas_int_l", "vas_lat_l", "med_gas_l", "lat_gas_l", "soleus_l",
                       "tib_post_l", "flex_dig_l", "flex_hal_l", "tib_ant_l", "per_brev_l", "per_long_l",
                       "per_tert_l", "ext_dig_l", "ext_hal_l", "ercspn_r", "ercspn_l", "intobl_r",
                       "intobl_l", "extobl_r", "extobl_l"]

        return action_spec

    @classmethod
    def get_default_xml_file_path(cls) -> str:
        """
        Returns the default path to the xml file of the environment.
        """
        return (loco_mujoco.PATH_TO_MODELS / "skeleton" / "skeleton_muscle.xml").as_posix()


class MjxSkeletonMuscle(SkeletonMuscle):

    """
    Mjx version of SkeletonMuscle.

    """

    mjx_enabled = True

    def __init__(self, timestep: float = 0.002,
                 n_substeps: int = 5,
                 **kwargs):
        """
        Constructor.

        Args:
            timestep (float): Timestep of the simulation.
            n_substeps (int): Number of substeps.
            **kwargs: Additional arguments.
        """
        if "model_option_conf" not in kwargs.keys():
            model_option_conf = dict(iterations=4, ls_iterations=8,
                                     disableflags=mujoco.mjtDisableBit.mjDSBL_EULERDAMP)
        else:
            model_option_conf = kwargs["model_option_conf"]
            del kwargs["model_option_conf"]
        super().__init__(timestep=timestep, n_substeps=n_substeps, model_option_conf=model_option_conf, **kwargs)


class HumanoidMuscle(SkeletonMuscle):
    """
    Wrapper class for SkeletonMuscle. Deprecated and will be removed in a future release.
    """
    def __init__(self, *args, **kwargs):
        warnings.warn(
            f"{self.__class__.__name__} is deprecated and will be removed in a future release. "
            f"Please use {super().__class__.__name__} instead.",
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*args, **kwargs)
